package no.statnett.ecp.pf;

import jakarta.jms.*;
import no.statnett.ecp.utils.Broker;

import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class MassProducer implements ArtemisClient {


    private final MessageProducer producer;
    private final Session session;
    private byte[] randomData;
    private int queueSize = -1; // This value is updated from the controller (Performance-class)
    private List<Integer> queueSizes = new ArrayList<>();
    private Change change;
    private int ttl;
    private boolean persistence;
    private Counter counter = new Counter("Prod");
    private Completion completion;
    private long sleep = 50; // default initial state
    private boolean regulate = true;


    public MassProducer(Map<String, String> amqpConfig, String queueName, int msgSizeB, int ttl, boolean persistence, Completion completion, boolean regulate) throws JMSException {
        Connection connection = Broker.createAndStartConnection(amqpConfig);
        session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
        producer = session.createProducer(session.createQueue(queueName));
        this.ttl = ttl;
        this.persistence = persistence;
        randomData = new byte[msgSizeB];
        new SecureRandom().nextBytes(randomData);
        this.completion = completion;
        this.regulate = regulate;
        if (!regulate) {
            this.sleep = 0;
        }
    }

    public void setQueueSize(int queueSize) {
        this.queueSize = queueSize;
        queueSizes.add(queueSize);
        if (queueSizes.size() > 50) {
            queueSizes.remove(0);
        }

        long queueSizeB = (long) queueSize * randomData.length;

        int firstQueueSize = queueSizes.get(0);
        int prevQueueSize = firstQueueSize;
        if (queueSizes.size() > 2)
            prevQueueSize = queueSizes.get(queueSizes.size() - 2);

        // Measure changes in queue size over long period (50 measurements, a measurement every 100ms)
        int queueChangeLong = queueSize - firstQueueSize;
        int queueChangeLongRate = queueSize > 0 ? (queueChangeLong * 100) / queueSize : queueChangeLong < 0 ? -100 : 0;

        // Measure changes in queue size from previous measurement (100ms), to respond more quickly if upward trend is quick
        int queueChangePrev = prevQueueSize - queueSize;
        int queueChangePrevRate = queueChangePrev > 0 ? (queueChangePrev * 100) / prevQueueSize : queueChangePrev < 0 ? -100 : 0;

        if (queueChangeLongRate == 0) {
            if (firstQueueSize == 0)
                change = Change.ZERO;
            else
                change = Change.FLAT;
        } else if (queueChangeLongRate < 0) {
            if (queueSizeB * 4 > Performance.maxQueueSizeB) {
                change = Change.DOWN_BUT_HIGH;
            } else if (queueSizeB * 12 > Performance.maxQueueSizeB) {
                change = Change.DOWN_BUT_MEDIUM;
            } else {
                change = Change.DOWN;
            }
        } else { // queuesize is increasing over long time, we must decide how to classify this
            if (queueChangeLongRate < 50 || queueSizeB * 10 < Performance.maxQueueSizeB) {
                if (queueChangePrevRate < 20)
                    change = Change.UP;
                else
                    change = Change.UP_MEDIUM;
            } else if (queueChangeLongRate < 100 || queueSizeB * 5 < Performance.maxQueueSizeB) {
                if (queueChangePrevRate < 20)
                    change = Change.UP_MEDIUM;
                else
                    change = Change.UP_FAST;
            } else {
                change = Change.UP_FAST;
            }
        }
//        System.out.println("Change is now " + change + " based on new queue size going from " + firstQueueSize + " to " + queueSize + ", a change of " + queueChangeLong + " (" + queueChangeLongRate + "% change)");
    }

    public void setMsgSizeKB(int msgSizeB) {
        this.randomData = new byte[msgSizeB];
        new SecureRandom().nextBytes(randomData);

    }

    public Counter getCounter() {
        return counter;
    }

    public long getSleep() {
        return sleep;
    }


    // We measure two metrics in order to determine the speed of our production (the production is measured as 'transmitted' (completion), not just adding to local buffer)
    // 1. QueueSize
    // 2. QueueSizeBytes
    // 2. Bytes produced
    // We aim at not exceeding maxQueueSize,

    @Override
    public void run() {
        long cycleStartTms = 0;
        int messageCount = 0;
        while (true) {
            try {
                long cycleTime = System.currentTimeMillis() - cycleStartTms;
                boolean initialization = queueSize < 0;
                boolean queueFull = ((long) queueSize * randomData.length) >= Performance.maxQueueSizeB;
                boolean bufferAboveLimit = completion.aboveLimit(Performance.bufferLimitB);
                if (cycleTime > 100 || initialization || queueFull || bufferAboveLimit) {
                    if (regulate) {
                        if (change == Change.ZERO) {
                            sleep -= 25;
                        } else if (change == Change.DOWN) {
                            sleep -= 5; // produce more
                        } else if (change == Change.DOWN_BUT_HIGH) {
                            sleep++; // produce less
                        } else if (change == Change.UP_FAST) {
                            sleep += 75;
                        } else if (change == Change.UP_MEDIUM) {
                            sleep += 50;
                        } else if (change == Change.UP) {
                            sleep += 20;
                        }
                        if (sleep > 100) {
                            sleep = 100;
                        } else if (sleep < 0) {
                            sleep = 0;
                        }
                        if (cycleTime > 100) {
                            cycleStartTms = System.currentTimeMillis(); // reset cycleTime immediately after dectecting end of previous cycle
                            counter.increase(cycleTime, messageCount, sleep);
                            messageCount = 0;
                        }
                        if (sleep > 0)
                            Thread.sleep(sleep);
                    } else {
                        if (cycleTime > 100) {
                            cycleStartTms = System.currentTimeMillis(); // reset cycleTime immediately after dectecting end of previous cycle
                            counter.increase(cycleTime, messageCount, 0);
                            messageCount = 0;
                        }
                    }
                }
                if (sleep < 100 && !initialization && !queueFull && !bufferAboveLimit) { // Will stop producing as soon as we are above certain limits
                    messageCount = produce(messageCount);
                }
            } catch (Throwable t) {
                System.out.println("Producer experienced an error " + t + ", but why quit? The producer will try again. But first sleep 30 seconds, to give the system the time to recover");
                try {
                    Thread.sleep(30000);
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    private int produce(int messageCount) throws JMSException {
        BytesMessage message = session.createBytesMessage();
        message.writeBytes(randomData);
        producer.send(message, persistence ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT, Message.DEFAULT_PRIORITY, ttl, completion);
        messageCount++;
        completion.messageSent();
        return messageCount;
    }
}
